
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "openglheader.h"

#include "utilities.h"
#include "lights.h"
#include "meshes.h"
#include "app3_a.h"

GLuint rprog_id[2];
GLuint rprog0_uvloc[2];
GLuint rprog1_uvloc[5];
static GLuint rshader_id[6];

GLuint trbi, trbuf, trbbp,
       lsbi, lsbuf, lsbbp;
GLuint empty_vao;
GLint  trbsize, trbofs[4],
       lsbsize, lsbofs[7];

void LoadMeshRenderingShaders ( void )
{
  static const GLchar *filename[] =
    { "app3_0.glsl.vert", "app3_0.glsl.geom", "app3_0.glsl.frag",
      "app3_1.glsl.vert", "app3_1.glsl.geom", "app3_1.glsl.frag" };
  static const GLuint shtype[] =
    { GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER,
      GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER };
  static const GLchar *UVnames[] =
    { "nsattr", "nv", "nhe", "nfac", "colour" };
  static const GLchar *UTBnames[] =
    { "TransBlock", "TransBlock.mm", "TransBlock.mmti", "TransBlock.vpm",
      "TransBlock.eyepos" };
  static const GLchar *ULSnames[] =
    { "LSBlock", "LSBlock.nls", "LSBlock.mask",
      "LSBlock.ls[0].ambient", "LSBlock.ls[0].direct", "LSBlock.ls[0].position",
      "LSBlock.ls[0].attenuation", "LSBlock.ls[1].ambient" };
  int i;

  for ( i = 0; i < 6; i++ )
    rshader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  rprog_id[0] = LinkShaderProgram ( 3, &rshader_id[0] );
  rprog_id[1] = LinkShaderProgram ( 3, &rshader_id[3] );
  rprog0_uvloc[0] = glGetUniformLocation ( rprog_id[0], UVnames[0] );
  rprog0_uvloc[1] = glGetUniformLocation ( rprog_id[0], UVnames[4] );
  for ( i = 0; i < 5; i++ )
    rprog1_uvloc[i] = glGetUniformLocation ( rprog_id[1], UVnames[i] );
  GetAccessToUniformBlock ( rprog_id[0], 4, &UTBnames[0],
                            &trbi, &trbsize, trbofs, &trbbp );
  AttachUniformBlockToBP ( rprog_id[1], UTBnames[0], trbbp );
  GetAccessToUniformBlock ( rprog_id[1], 7, &ULSnames[0],
                            &lsbi, &lsbsize, lsbofs, &lsbbp );
  trbuf = NewUniformBlockObject ( trbsize, trbbp );
  lsbuf = NewUniformBlockObject ( lsbsize, lsbbp );
  glGenVertexArrays ( 1, &empty_vao );
  glBindVertexArray ( empty_vao );
  glVertexAttrib1f ( 0, 0.0 );
  ExitIfGLError ( "LoadMeshRenderingShaders" );
} /*LoadMeshRenderingShaders*/

void DestroyMeshRenderingShaders ( void )
{
  int i;

  glUseProgram ( 0 );
  for ( i = 0; i < 6; i++ )
    glDeleteShader ( rshader_id[i] );
  for ( i = 0; i < 2; i++ )
    glDeleteProgram ( rprog_id[i] );
  glDeleteVertexArrays ( 1, &empty_vao );
  glDeleteBuffers ( 1, &trbuf );
  glDeleteBuffers ( 1, &lsbuf );
  ExitIfGLError ( "DestroyMeshRenderingShaders" );
} /*DestroyMeshRenderingShaders*/

void DrawMeshEdges ( GPUmesh *mesh, GLfloat colour[4] )
{
  glUseProgram ( rprog_id[0] );
  glUniform4fv ( rprog0_uvloc[1], 1, colour );
  glUniform1i  ( rprog0_uvloc[0], mesh->nsattr );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 0, mesh->mbuf[0] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 1, mesh->mbuf[1] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 2, mesh->mbuf[2] );
  glBindVertexArray ( empty_vao );
  glDrawArraysInstanced ( GL_LINES, 0, 2, mesh->nhe );
  ExitIfGLError ( "DrawMeshEdges" );
} /*DrawMeshEdges*/

void DrawMeshFacets ( GPUmesh *mesh, GLfloat colour[4] )
{
  glUseProgram ( rprog_id[1] );
  glUniform4fv ( rprog1_uvloc[4], 1, colour );
  glUniform1i ( rprog1_uvloc[0], mesh->nsattr );
  glUniform1i ( rprog1_uvloc[1], mesh->nv );
  glUniform1i ( rprog1_uvloc[2], mesh->nhe );
  glUniform1i ( rprog1_uvloc[3], mesh->nfac );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 0, mesh->mbuf[0] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 1, mesh->mbuf[1] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 2, mesh->mbuf[2] );
  glBindVertexArray ( empty_vao );
  glDrawArraysInstanced ( GL_TRIANGLE_FAN, 0, 4, mesh->nfac );
  ExitIfGLError ( "DrawMeshFacets" );
} /*DrawMeshFacets*/

