
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "openglheader.h"

#include "utilities.h"
#include "meshes.h"

#define MVFBUF   mbuf[0]
#define MHEBUF   mbuf[1]
#define VCBUF    mbuf[2]

#define FHEMASK  0x01FFFFFF
#define TAGMASK  0x02000000
#define DEGMASK  0xFC000000
#define DEGSHIFT         26

#define SSB GL_SHADER_STORAGE_BUFFER

/* ////////////////////////////////////////////////////////////////////////// */
char ReallocGPUmesh ( GPUmesh *gmesh, int nv, int nhe, int nfac, int nsattr,
                      int pdim, int pofs, int nvofs, int tdim, int tofs )
{
  int i;

  for ( i = 0; i < 3; i++ )
    if ( gmesh->mbuf[i] > 0 ) glDeleteBuffers ( 1, &gmesh->mbuf[i] );
  glGenBuffers ( 3, gmesh->mbuf );
  glBindBuffer ( SSB, gmesh->MVFBUF );
  glBufferData ( SSB, (nv+nfac+2*nhe)*sizeof(GLuint), NULL, GL_DYNAMIC_DRAW );
  glBindBuffer ( SSB, gmesh->MHEBUF );
  glBufferData ( SSB, nhe*4*sizeof(GLint), NULL, GL_DYNAMIC_DRAW );
  glBindBuffer ( SSB, gmesh->VCBUF );
  glBufferData ( SSB, nv*nsattr*sizeof(GLfloat), NULL, GL_DYNAMIC_DRAW );
  gmesh->nsattr = nsattr;  gmesh->pdim = pdim;  gmesh->pofs = pofs;
  gmesh->nvofs = nvofs;  gmesh->tdim = tdim;  gmesh->tofs = tofs;
  gmesh->nv = nv;
  gmesh->nhe = nhe;
  gmesh->nfac = nfac;
  ExitIfGLError ( "ReallocGPUmesh" );
  return true;
} /*ReallocGPUmesh*/

void FreeCPUmeshTab ( CPUmesh *cmesh )
{
  if ( cmesh->mv ) free ( cmesh->mv );
  if ( cmesh->vc ) free ( cmesh->vc );
  if ( cmesh->mhe ) free ( cmesh->mhe );
  if ( cmesh->mfac ) free ( cmesh->mfac );
  if ( cmesh->mvhei ) free ( cmesh->mvhei );
  if ( cmesh->mfhei ) free ( cmesh->mfhei );
  memset ( cmesh, 0, sizeof(CPUmesh) );
} /*FreeCPUmeshTab*/

char ReallocCPUmesh ( CPUmesh *cmesh, int nv, int nhe, int nfac, int nsattr,
                      int pdim, int pofs, int nvofs, int tdim, int tofs )
{
  FreeCPUmeshTab ( cmesh );
  cmesh->mv = malloc ( nv*sizeof(BSMvertex) );
  cmesh->vc = malloc ( nv*nsattr*sizeof(double) );
  cmesh->mhe = malloc ( nhe*sizeof(BSMhalfedge) );
  cmesh->mfac = malloc ( nfac*sizeof(BSMfacet) );
  cmesh->mvhei = malloc ( nhe*sizeof(int) );
  cmesh->mfhei = malloc ( nhe*sizeof(int) );
  if ( !cmesh->mv | !cmesh->vc | !cmesh->mhe | !cmesh->mfac |
       !cmesh->mvhei | !cmesh->mfhei ) {
    FreeCPUmeshTab ( cmesh );
    return false;
  }
  else {
    cmesh->nsattr = nsattr;  cmesh->pdim = pdim;  cmesh->pofs = pofs;
    cmesh->nvofs = nvofs;  cmesh->tdim = tdim;  cmesh->tofs = tofs;  
    cmesh->nv = nv;
    cmesh->nhe = nhe;
    cmesh->nfac = nfac;

memset ( cmesh->mv, 0, nv*sizeof(BSMvertex) );
memset ( cmesh->vc, 0, nv*nsattr*sizeof(double) );
memset ( cmesh->mhe, 0, nhe*sizeof(BSMhalfedge) );
memset ( cmesh->mfac, 0, nfac*sizeof(BSMfacet) );
memset ( cmesh->mvhei, 0, nhe*sizeof(int) );
memset ( cmesh->mfhei, 0, nhe*sizeof(int) );

    return true;
  }
} /*ReallocCPUmesh*/

char CPUmeshToGPU ( CPUmesh *cmesh, GPUmesh *gmesh )
{
  int       dim, nv, nhe, nfac;
  BSMvertex *mv;
  BSMfacet  *mfac;
  GLuint    *vf;
  GLfloat   *vc;
  int       i, size;

  if ( ReallocGPUmesh ( gmesh, nv = cmesh->nv, nhe = cmesh->nhe,
                        nfac = cmesh->nfac, dim = cmesh->nsattr, cmesh->pdim,
                        cmesh->pofs, cmesh->nvofs, cmesh->tdim, cmesh->tofs ) ) {
    size = (nv > nfac ? nv : nfac)*sizeof(GLuint);
    i = nv*dim*sizeof(GLfloat);
    if ( size < i ) size = i;
    if ( !(vf = (GLuint*)malloc ( size )) )
      goto failure;
    glBindBuffer ( SSB, gmesh->MVFBUF );
    mv = cmesh->mv;  mfac = cmesh->mfac;
    for ( i = 0; i < nv; i++ )
      vf[i] = mv[i].firsthalfedge + (mv[i].degree << DEGSHIFT);
    glBufferSubData ( SSB, 0, nv*sizeof(GLuint), vf );
    for ( i = 0; i < nfac; i++ )
      vf[i] = mfac[i].firsthalfedge + (mfac[i].degree << DEGSHIFT);
    glBufferSubData ( SSB, nv*sizeof(GLuint), nfac*sizeof(GLuint), vf );
    glBufferSubData ( SSB, (nv+nfac)*sizeof(GLuint), nhe*sizeof(GLint),
                      cmesh->mvhei );
    glBufferSubData ( SSB, (nv+nfac+nhe)*sizeof(GLuint), nhe*sizeof(GLint),
                      cmesh->mfhei );
    glBindBuffer ( SSB, gmesh->MHEBUF );
    glBufferSubData ( SSB, 0, nhe*4*sizeof(GLint), cmesh->mhe );
    vc = (GLfloat*)vf;
    for ( i = 0; i < nv*dim; i++ )
      vc[i] = (GLfloat)cmesh->vc[i];
    glBindBuffer ( SSB, gmesh->VCBUF );
    glBufferSubData ( SSB, 0, nv*dim*sizeof(GLfloat), vc );
    free ( vf );
    ExitIfGLError ( "CPUmeshToGPU" );
    return true;
  }
failure:
  return false;
} /*CPUmeshToGPU*/

char GPUmeshToCPU ( GPUmesh *gmesh, CPUmesh *cmesh )
{
  int       dim, nv, nhe, nfac;
  BSMvertex *mv;
  BSMfacet  *mfac;
  GLuint    *vf;
  GLfloat   *vc;
  int       i, size;
  
  if ( ReallocCPUmesh ( cmesh, nv = gmesh->nv, nhe = gmesh->nhe,
                        nfac = gmesh->nfac, dim = gmesh->nsattr, gmesh->pdim,
                        gmesh->pofs, cmesh->nvofs, cmesh->tdim, cmesh->tofs ) ) {
    size = (nv > nfac ? nv : nfac)*sizeof(GLuint);
    i = nv*dim*sizeof(GLfloat);
    if ( size < i ) size = i;
    if ( !(vf = (GLuint*)malloc ( size )) )
      goto failure;
    mv = cmesh->mv;  mfac = cmesh->mfac;
    glBindBuffer ( SSB, gmesh->MVFBUF );
    glGetBufferSubData ( SSB, 0, nv*sizeof(GLuint), vf );
    for ( i = 0; i < nv; i++ ) {
      mv[i].firsthalfedge = vf[i] & FHEMASK;
      mv[i].degree = vf[i] >> DEGSHIFT;
    }
    glGetBufferSubData ( SSB, nv*sizeof(GLint), nfac*sizeof(GLuint), vf );
    for ( i = 0; i < nfac; i++ ) {
      mfac[i].firsthalfedge = vf[i] & FHEMASK;
      mfac[i].degree = vf[i] >> DEGSHIFT;
    }
    glGetBufferSubData ( SSB, (nv+nfac)*sizeof(GLint), nhe*sizeof(GLint),
                         cmesh->mvhei );
    glGetBufferSubData ( SSB, (nv+nfac+nhe)*sizeof(GLint), nhe*sizeof(GLint),
                         cmesh->mfhei );
    glBindBuffer ( SSB, gmesh->MHEBUF );
    glGetBufferSubData ( SSB, 0, nhe*4*sizeof(GLint), cmesh->mhe );
    vc = (GLfloat*)vf;
    glBindBuffer ( SSB, gmesh->VCBUF );
    glGetBufferSubData ( SSB, 0, nv*dim*sizeof(GLfloat), vc );
    for ( i = 0; i < nv*dim; i++ )
      cmesh->vc[i] = (double)vc[i];
    free ( vf );
    ExitIfGLError ( "GPUmeshToCPU" );
    return true;
  }
failure:
  return false;
} /*GPUmeshToCPU*/

void DeleteCPUmesh ( CPUmesh *cmesh )
{
  if ( cmesh->mv ) free ( cmesh->mv );
  if ( cmesh->mhe ) free ( cmesh->mhe );
  if ( cmesh->mfac ) free ( cmesh->mfac );
  if ( cmesh->mvhei ) free ( cmesh->mvhei );
  if ( cmesh->mfhei ) free ( cmesh->mfhei );
  if ( cmesh->vc ) free ( cmesh->vc );
  free ( cmesh );
} /*DeleteCPUmesh*/

void DeleteGPUmesh ( GPUmesh *gmesh )
{
  glDeleteBuffers ( 3, gmesh->mbuf );
  free ( gmesh );
} /*DeleteGPUmesh*/

