
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "openglheader.h"

#include "utilities.h"
#include "lights.h"
#include "bezpatches.h"
#include "meshes.h"
#include "linkage.h"
#include "app3c_a.h"

static GLuint rshader_id[13];
static GLuint rprog_id[6];
static GLuint rprog0_uvloc[4];
static GLuint rprog1_uvloc[8];
static GLuint rprog2_uvloc[7];
static GLuint LightProcInd1, LambertProcInd1, BlinnPhongProcInd1;
static GLuint LightProcInd4, LambertProcInd4, BlinnPhongProcInd4;
static GLint  LightProcLoc1, LightProcLoc4, BezNormalsLoc4, BezTessLevelLoc4,
              ColourSourceLoc4, BezTessLevelLoc5, ColourLoc3;

GLuint trbi, trbuf, trbbp, lsbi, lsbuf, lsbbp;
GLint  trbsize, trbofs[6], lsbsize, lsbofs[9];

static GLuint LinkMyShaderProgram ( const int *shn, const GLuint *shaders )
{
  GLuint sh[5], i;

  for ( i = 0; i < shn[0]; i++ )
    sh[i] = shaders[shn[i+1]];
  return LinkShaderProgram ( shn[0], sh );
} /*LinkMyShaderProgram*/

void LoadRenderingShaders ( void )
{
  static const GLchar *filename[] =
    { "app3C0.glsl.vert", "app3C1.glsl.vert", "app3C2.glsl.vert", "app3C3.glsl.vert",
      "app3C4.glsl.vert", "app3C0.glsl.tesc", "app3C0.glsl.tese", "app3C1.glsl.tese",
      "app3C0.glsl.geom", "app3C1.glsl.geom", "app3C2.glsl.geom", "app3C0.glsl.frag",
      "app3C1.glsl.frag" };
  static const GLuint shtype[] =
    { GL_VERTEX_SHADER, GL_VERTEX_SHADER, GL_VERTEX_SHADER, GL_VERTEX_SHADER,
      GL_VERTEX_SHADER, GL_TESS_CONTROL_SHADER, GL_TESS_EVALUATION_SHADER,
      GL_TESS_EVALUATION_SHADER, GL_GEOMETRY_SHADER, GL_GEOMETRY_SHADER,
      GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER, GL_FRAGMENT_SHADER };
  static const int p0sh[] = {3,0,8,11};
  static const int p1sh[] = {3,1,9,12};
  static const int p2sh[] = {2,2,11};
  static const int p3sh[] = {2,4,11};
  static const int p4sh[] = {5,3,5,6,10,12};
  static const int p5sh[] = {4,3,5,7,11};
  static const GLchar *UVnames[] =
    { "nsattr", "pdim", "pofs", "nv", "nhe", "nfac", "nvofs", "MeshNormals",
      "colour", "BezNormals", "BezTessLevel", "ColourSource" };
  static const GLchar *UTBnames[] =
    { "TransBlock", "TransBlock.mm", "TransBlock.mmti", "TransBlock.vm",
      "TransBlock.pm", "TransBlock.vpm", "TransBlock.eyepos" };
  static const GLchar *ULSnames[] =
    { "LSBlock", "LSBlock.nls", "LSBlock.mask", "LSBlock.shmask",
      "LSBlock.ls[0].ambient", "LSBlock.ls[0].direct", "LSBlock.ls[0].position",
      "LSBlock.ls[0].attenuation", "LSBlock.ls[0].shadow_vpm",
      "LSBlock.ls[1].ambient" };
  static const GLchar *UMatNames[] =
    { "MatBlock", "MatBlock.mat.ambref", "MatBlock.mat.dirref",
      "MatBlock.mat.specref", "MatBlock.mat.shininess",
      "MatBlock.mat.wa", "MatBlock.mat.we" };
  static const GLchar *UCPnames[] =
    { "CPoints", "CPoints.cp" };
  static const GLchar *UCPInames[] = { "CPIndices", "CPIndices.cpi" };
  static const GLchar *UBezPatchNames[] =
    { "BezPatch", "BezPatch.dim", "BezPatch.udeg", "BezPatch.vdeg",
      "BezPatch.stride_u", "BezPatch.stride_v", "BezPatch.stride_p",
      "BezPatch.stride_q", "BezPatch.nq", "BezPatch.use_ind",
      "BezPatch.Colour", "BezPatch.TessLevel", "BezPatch.BezNormals" };
  static const GLchar LightProcName[] = "Lighting";
  static const GLchar LambertProcName[] = "LambertLighting";
  static const GLchar BlinnPhongProcName[] = "BlinnPhongLighting";

  int i;

  for ( i = 0; i < 13; i++ )
    rshader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  rprog_id[0] = LinkMyShaderProgram ( p0sh, rshader_id );
  rprog_id[1] = LinkMyShaderProgram ( p1sh, rshader_id );
  rprog_id[2] = LinkMyShaderProgram ( p2sh, rshader_id );
  rprog_id[3] = LinkMyShaderProgram ( p3sh, rshader_id );
  rprog_id[4] = LinkMyShaderProgram ( p4sh, rshader_id );
  rprog_id[5] = LinkMyShaderProgram ( p5sh, rshader_id );
  for ( i = 0; i < 3; i++ )
    rprog0_uvloc[i] = glGetUniformLocation ( rprog_id[0], UVnames[i] );
  rprog0_uvloc[3] = glGetUniformLocation ( rprog_id[0], UVnames[8] );
  for ( i = 0; i < 8; i++ )
    rprog1_uvloc[i] = glGetUniformLocation ( rprog_id[1], UVnames[i] );
  for ( i = 0; i < 6; i++ )
    rprog2_uvloc[i] = glGetUniformLocation ( rprog_id[2], UVnames[i] );
  rprog2_uvloc[6] = glGetUniformLocation ( rprog_id[2], UVnames[8] );
  LightProcLoc1 = glGetSubroutineUniformLocation ( rprog_id[1],
                         GL_FRAGMENT_SHADER, LightProcName );
  LambertProcInd1 = glGetSubroutineIndex ( rprog_id[1], GL_FRAGMENT_SHADER,
                         LambertProcName );
  BlinnPhongProcInd1 = glGetSubroutineIndex ( rprog_id[1], GL_FRAGMENT_SHADER,
                         BlinnPhongProcName );
  LightProcInd1 = BlinnPhongProcInd1;
  LightProcLoc4 = glGetSubroutineUniformLocation ( rprog_id[4],
                         GL_FRAGMENT_SHADER, LightProcName );
  LambertProcInd4 = glGetSubroutineIndex ( rprog_id[4], GL_FRAGMENT_SHADER,
                         LambertProcName );
  BlinnPhongProcInd4 = glGetSubroutineIndex ( rprog_id[4], GL_FRAGMENT_SHADER,
                         BlinnPhongProcName );
  LightProcInd4 = BlinnPhongProcInd4;
  GetAccessToUniformBlock ( rprog_id[0], 6, &UTBnames[0],
                            &trbi, &trbsize, trbofs, &trbbp );
  for ( i = 1; i < 6; i++ )
    AttachUniformBlockToBP ( rprog_id[i], UTBnames[0], trbbp );
  GetAccessToUniformBlock ( rprog_id[1], 9, &ULSnames[0],
                            &lsbi, &lsbsize, lsbofs, &lsbbp );
  GetAccessToUniformBlock ( rprog_id[1], 6, &UMatNames[0],
                            &matbi, &matbsize, matbofs, &matbbp );
  GetAccessToUniformBlock ( rprog_id[4], 1, &UCPnames[0],
                            &cpbi, &i, cpbofs, &cpbbp );
  GetAccessToUniformBlock ( rprog_id[4], 1, &UCPInames[0],
                            &cpibi, &i, cpibofs, &cpibbp );
  GetAccessToUniformBlock ( rprog_id[4], 12, &UBezPatchNames[0],
                            &bezpbi, &bezpbsize, bezpbofs, &bezpbbp );
  AttachUniformBlockToBP ( rprog_id[4], ULSnames[0], lsbbp );
  AttachUniformBlockToBP ( rprog_id[4], UMatNames[0], matbbp );
  AttachUniformBlockToBP ( rprog_id[5], UCPnames[0], cpbbp );
  AttachUniformBlockToBP ( rprog_id[5], UCPInames[0], cpibbp );
  AttachUniformBlockToBP ( rprog_id[5], UBezPatchNames[0], bezpbbp );
  AttachUniformBlockToBP ( rprog_id[3], UCPnames[0], cpbbp );
  AttachUniformBlockToBP ( rprog_id[3], UCPInames[0], cpibbp );
  AttachUniformBlockToBP ( rprog_id[3], UBezPatchNames[0], bezpbbp );
  BezNormalsLoc4 = glGetUniformLocation ( rprog_id[4], UVnames[9] );
  BezTessLevelLoc4 = glGetUniformLocation ( rprog_id[4], UVnames[10] );
  ColourSourceLoc4 = glGetUniformLocation ( rprog_id[4], UVnames[11] );
  BezTessLevelLoc5 = glGetUniformLocation ( rprog_id[5], UVnames[10] );
  ColourLoc3 = glGetUniformLocation ( rprog_id[3], UVnames[8] );
  trbuf = NewUniformBlockObject ( trbsize, trbbp );
  lsbuf = NewUniformBlockObject ( lsbsize, lsbbp );
  matbuf[1] = NewUniformBlockObject ( matbsize, matbbp );
  matbuf[0] = NewUniformBlockObject ( matbsize, matbbp );
  ConstructEmptyVAO ();
  ExitIfGLError ( "LoadMeshRenderingShaders" );
} /*LoadRenderingShaders*/

void DestroyRenderingShaders ( void )
{
  int i;

  glUseProgram ( 0 );
  for ( i = 0; i < 13; i++ )
    glDeleteShader ( rshader_id[i] );
  for ( i = 0; i < 6; i++ )
    glDeleteProgram ( rprog_id[i] );
  DeleteEmptyVAO ();
  glDeleteBuffers ( 1, &trbuf );
  glDeleteBuffers ( 1, &lsbuf );
  glDeleteBuffers ( 2, matbuf );
  ExitIfGLError ( "DestroyMeshRenderingShaders" );
} /*DestroyRenderingShaders*/

void DrawMeshEdges ( GPUmesh *mesh, const GLfloat colour[4] )
{
  glUseProgram ( rprog_id[0] );
  glUniform4fv ( rprog0_uvloc[3], 1, colour );
  glUniform1i  ( rprog0_uvloc[0], mesh->nsattr );
  glUniform1i  ( rprog0_uvloc[1], mesh->pdim );
  glUniform1i  ( rprog0_uvloc[2], mesh->pofs );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 0, mesh->mbuf[0] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 1, mesh->mbuf[1] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 2, mesh->mbuf[2] );
  glBindVertexArray ( empty_vao );
  glDrawArraysInstanced ( GL_LINES, 0, 2, mesh->nhe );
  ExitIfGLError ( "DrawMeshEdges" );
} /*DrawMeshEdges*/

void DrawMeshFacets ( GPUmesh *mesh, int mat, char nvectors, char final )
{
  if ( final ) {
    glUseProgram ( rprog_id[1] );
    glUniformSubroutinesuiv ( GL_FRAGMENT_SHADER, 1, &LightProcInd1 );
    glUniform1i ( rprog1_uvloc[0], mesh->nsattr );
    glUniform1i ( rprog1_uvloc[1], mesh->pdim );
    glUniform1i ( rprog1_uvloc[2], mesh->pofs );
    glUniform1i ( rprog1_uvloc[3], mesh->nv );
    glUniform1i ( rprog1_uvloc[4], mesh->nhe );
    glUniform1i ( rprog1_uvloc[5], mesh->nfac );
    glUniform1i ( rprog1_uvloc[6], mesh->nvofs );
    glUniform1i ( rprog1_uvloc[7], (int)(nvectors && (mesh->nvofs >= 0)) );
    ChooseMaterial ( mat );
  }
  else {
    glUseProgram ( rprog_id[2] );
    glUniform1i ( rprog2_uvloc[0], mesh->nsattr );
    glUniform1i ( rprog2_uvloc[1], mesh->pdim );
    glUniform1i ( rprog2_uvloc[2], mesh->pofs );
    glUniform1i ( rprog2_uvloc[3], mesh->nv );
    glUniform1i ( rprog2_uvloc[4], mesh->nhe );
    glUniform1i ( rprog2_uvloc[5], mesh->nfac );
  }
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 0, mesh->mbuf[0] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 1, mesh->mbuf[1] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 2, mesh->mbuf[2] );
  glBindVertexArray ( empty_vao );
  glDrawArraysInstanced ( GL_TRIANGLE_FAN, 0, 4, mesh->nfac );
  ExitIfGLError ( "DrawMeshFacets" );
} /*DrawMeshFacets*/

void DrawBezPatches ( BezierPatchObjf *bp, int lod, char final )
{
  int TessLevel[4] = {4,8,16,32};

  if ( bp && lod > 0 && lod <= 4 ) {
    if ( final ) {
      glUseProgram ( rprog_id[4] );
      glUniform1i ( BezTessLevelLoc4, TessLevel[lod-1] );
      glUniformSubroutinesuiv ( GL_FRAGMENT_SHADER, 1, &LightProcInd4 );
    }
    else {
      glUseProgram ( rprog_id[5] );
      glUniform1i ( BezTessLevelLoc5, TessLevel[lod-1] );
    }
    DrawBezierPatches ( bp );
  }
} /*DrawBezPatches*/

void DrawBezPNets ( BezierPatchObjf *bp, const GLfloat colour[4] )
{
  glUseProgram ( rprog_id[3] );
  glUniform4fv ( ColourLoc3, 1, colour );
  DrawBezierNets ( bp );
} /*DrawBezPNets*/

