#version 450 core

#define MAX_NLIGHTS 8

layout(early_fragment_tests) in;

in NVertex {
  vec3 Position;
  vec3 snv, tnv;
  vec4 ShadowPos[MAX_NLIGHTS];
} In;

out vec4 out_Colour;

uniform TransBlock {
  mat4 mm, mmti, vm, pm, vpm;
  vec4 eyepos;
} trb;

struct LSPar {
    vec4 ambient;
    vec4 direct;
    vec4 position;
    vec3 attenuation;
    mat4 shadow_vpm;
  };

uniform LSBlock {
    uint  nls;              /* liczba zrodel swiatla */
    uint  mask;             /* maska wlaczonych zrodel */
    uint  shmask;           /* maska tekstur cienia */
    LSPar ls[MAX_NLIGHTS];  /* poszczegolne zrodla swiatla */
  } light;

struct Material {
    vec4  ambref;     /* odbijanie swiatla rozproszonego */
    vec4  dirref;     /* odbijanie swiatla kierunkowego */
    vec4  specref;    /* odbijanie zwierciadlane swiatla */
    float shininess, wa, we;  /* polysk */
  };

uniform MatBlock {
    Material mat;
  } mat;

layout (binding = 0) uniform sampler2DShadow shtex[MAX_NLIGHTS];

subroutine void LightingProc ( void );

subroutine uniform LightingProc Lighting;

Material mm;
vec3     normal, nv;

vec3 posDifference ( vec4 p, vec3 pos, out float dist )
{
  vec3 v;

  if ( p.w != 0.0 ) {
    v = p.xyz/p.w-pos.xyz;
    dist = sqrt ( dot ( v, v ) );
  }
  else
    v = p.xyz;
  return normalize ( v );
} /*posDifference*/

float attFactor ( vec3 att, float dist )
{
  return 1.0/(((att.z*dist)+att.y)*dist+att.x);
} /*attFactor*/

float IsEnlighted ( uint l )
{
  return textureProj ( shtex[l], In.ShadowPos[l] );
} /*IsEnlighted*/

subroutine (LightingProc) void LambertLighting ( void )
{
  vec3  lv, vv;
  float d, e, s, dist;
  uint  i, mask;

  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, nv );
  out_Colour = vec4(0.0);
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      out_Colour += light.ls[i].ambient * mm.ambref;
      s = ((light.shmask & mask) != 0) ? IsEnlighted ( i ) : 1.0;
      if ( s > 0.0 ) {
        lv = posDifference ( light.ls[i].position, In.Position, dist );
        d = dot ( lv, normal );
        if ( e > 0.0 ) {
          if ( dot ( lv, nv ) > 0.0 ) {
            if ( light.ls[i].position.w != 0.0 )
              d *= attFactor ( light.ls[i].attenuation, dist );
            out_Colour += (s*d * light.ls[i].direct) * mm.dirref;
          }
        }
        else {
          if ( dot ( lv, nv ) < 0.0 ) {
            if ( light.ls[i].position.w != 0.0 )
              d *= attFactor ( light.ls[i].attenuation, dist );
            out_Colour -= (s*d * light.ls[i].direct) * mm.dirref;
          }
        }
      }
    }
  out_Colour = vec4 ( clamp ( out_Colour.rgb, 0.0, 1.0 ), 1.0 );
} /*LambertLighting*/

float wFactor ( float lvn, float wa, float we )
{
  return 1.0+(wa-1.0)*pow ( 1.0-lvn, we );
} /*wFactor*/

subroutine (LightingProc) void BlinnPhongLighting ( void )
{
  vec3  lv, vv, hv;
  float a, d, e, f, s, dist;
  uint  i, mask;

  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, nv );
  out_Colour = vec4(0.0);
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      out_Colour += light.ls[i].ambient * mm.ambref;
      s = ((light.shmask & mask) != 0) ? IsEnlighted ( i ) : 1.0;
      if ( s > 0.0 ) {
        lv = posDifference ( light.ls[i].position, In.Position, dist );
        d = dot ( lv, normal );
        if ( e > 0.0 ) {
          if ( dot ( lv, nv ) > 0.0 ) {
            a = light.ls[i].position.w != 0.0 ?
                  s*attFactor ( light.ls[i].attenuation, dist ) : s;
            out_Colour += (a*d*light.ls[i].direct) * mm.dirref;
            hv = normalize ( lv+vv );
            f = pow ( dot ( hv, normal ), mm.shininess );
            out_Colour += (a*f*wFactor(d,mm.wa,mm.we))*mm.specref;
          }
        }
        else {
          if ( dot ( lv, nv ) < 0.0 ) {
            a = light.ls[i].position.w != 0.0 ?
                  s*attFactor ( light.ls[i].attenuation, dist ) : s;
            out_Colour -= (a*d*light.ls[i].direct) * mm.dirref;
            hv = normalize ( lv+vv );
            f = pow ( -dot ( hv, normal ), mm.shininess );
            out_Colour += (a*f*wFactor(-d,mm.wa,mm.we))*mm.specref;
          }
        }
      }
    }
  out_Colour = vec4 ( clamp ( out_Colour.rgb, 0.0, 1.0 ), 1.0 );
} /*BlinnPhongLighting*/

void main ( void )
{
  nv = In.tnv;
  normal = normalize ( In.snv );
  mm = mat.mat;
  Lighting ();
} /*main*/

