#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "openglheader.h"
#include <GL/glx.h>

#include "utilities.h"
#include "xvariables.h"
#include "xwidgets.h"
#include "xknotswidget.h"
#include "initglxctx.h"
#include "bezpatches.h"
#include "meshes.h"
#include "linkage.h"
#include "lights.h"
#include "app3d_a.h"
#include "app3d_b.h"

Window        window[4];
GC            xgc;
Atom          aAnimate;
char          terminate;

int           window0_width, window0_height;
xwinmenu      *wm1, *wm2, *wm3;
xwidget       *mywdg, *playbtn;
xKnotsWidgetf *myknotwdg;
myglwindata   myglwdata;
mypalmscene   myscene;
char          opti = 0;

char str_EXIT[]    = "Exit";
char str_SET[]     = "Set";
char str_LEFT[]    = "<-";
char str_RIGHT[]   = "->";
char str_PANZOOM[] = "pan/zoom";
char str_EDIT[]    = "edit";

void Win1Callback ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  switch ( msg ) {
case WDGMSG_BUTTON_PRESS:
    switch ( wdg->id ) {
  case BTN_ID_EXIT:
      terminate = true;
      break;
  default:
      break;
    }
    break;

case WDGMSG_SWITCH_CHANGE:
    switch ( wdg->id ) {
  case SW_ID_MESH1:  case SW_ID_MESH2:  case SW_ID_MESH3:  case SW_ID_MESH4:
      if ( myscene.meshsw[wdg->id-SW_ID_MESH0] ) {
        memset ( &myscene.meshsw[1], false, NPALMMESHES );
        myscene.meshsw[(int)(myscene.lod = wdg->id-SW_ID_MESH0)] = true;
        kl_Articulate ( myscene.lkg );
      }
      else
        myscene.lod = -1;
/*printf ( "lod = %d\n", (int)myscene.lod );*/
      goto win2changed;
  case SW_ID_MESH0:
      goto win2changed;
  default:
      break;
    }

case WDGMSG_SLIDEBAR_CHANGE:
    if ( wdg->id >= SL_ID_ARTP0 && wdg->id < SL_ID_ARTP0+NKLARTPARAMS ) {
/*printf ( "param %d, %f\n", wdg->id-SL_ID_ARTP0, palmartp[wdg->id-SL_ID_ARTP0] );*/
      SetArticulationParameter ( &myscene, wdg->id-SL_ID_ARTP0 );
      kl_Articulate ( myscene.lkg );
win2changed:
      wm2->changed = true;
      PostMenuExposeEvent ( wm2 );
      break;
    }
    break;

default:
    break;
  }
} /*Win1CallBack*/

xwinmenu *SetupApp3dMenu ( void )
{
  xwinmenu *wm;
  int      i;

  if ( !(wm = NewWinMenu ( window[1], MENU1_WIDTH, WIN0_HEIGHT-MENU3_HEIGHT,
                           0, 0, NULL, NULL, Win1Callback )) )
    ExitOnError ( "SetupApp3dMenu" );
  NewButton ( wm, BTN_ID_EXIT, 60, 18, 2, 2, str_EXIT );
  for ( i = 0; i <= NPALMMESHES; i++ )
    NewSwitch ( wm, SW_ID_MESH0+i, 16, 16, 2+20*i, 22, NULL, &myscene.meshsw[i] );
  for ( i = 0; i < NKLARTPARAMS; i++ )
    NewSlidebarf ( wm, SL_ID_ARTP0+i, 116, 10, 2, 46+15*i, &myscene.palmartp[i] );
  return wm;
} /*SetupApp3dMenu*/

/* ////////////////////////////////////////////////////////////////////////// */
void Win0ConfigureNotify ( int width, int height )
{
  window0_width  = width;
  window0_height = height;
  XMoveResizeWindow ( xdisplay, window[1], 0, 0,
                      MENU1_WIDTH, height-MENU3_HEIGHT );
  XMoveResizeWindow ( xdisplay, window[2], MENU1_WIDTH, 0,
                      width-MENU1_WIDTH, height-MENU3_HEIGHT );
  XMoveResizeWindow ( xdisplay, window[3], 0, height-MENU3_HEIGHT,
                      width, MENU3_HEIGHT );
} /*Win0ConfigureNotify*/

void Win0MessageProc ( XEvent *ev )
{
  switch ( ev->xany.type ) {
case ConfigureNotify:
    Win0ConfigureNotify ( ev->xconfigure.width, ev->xconfigure.height );
    break;
default:
    break;
  }
} /*Win0MessageProc*/

/* ////////////////////////////////////////////////////////////////////////// */
void My3DWidgetRedraw ( struct xwidget *wdg )
{
  DrawSceneToWindow ( &myscene, (myglwindata*)wdg->data0, wdg->r.width, wdg->r.height );
} /*My3DWidgetRedraw*/

void My3DWidgetReshape ( xwidget *wdg, short w, short h )
{
  myglwindata *wd;

  wdg->r.width = w;  wdg->r.height = h;
  wd = (myglwindata*)wdg->data0;
  wd->bottom = -(wd->top = 0.5533);
  wd->left = -(wd->right = 0.5533*(float)w/(float)h );
  wd->near = 5.0;  wd->far = 15.0;
  wd->rl = wd->right - wd->left;  wd->tb = wd->top - wd->bottom;
  M4x4Frustumf ( wd->pm, NULL, wd->left, wd->right, wd->bottom, wd->top,
                 wd->near, wd->far );
  SetVPMatrix ( wd->vm, wd->pm, wd->eyepos );
} /*My3DWidgetReshape*/

char My3DWidgetInput ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  myglwindata *wd;

  wd = (myglwindata*)wdg->data0;
  switch ( wdg->state ) {
case WDGSTATE_DEFAULT:
    switch ( msg ) {
  case XWMSG_BUTTON_PRESS:
      if ( key == Button1 ) {
        wd->prevx = x;  wd->prevy = y;
        wdg->state = WDGSTATE_VIEW_TURNING;
        GrabInput ( wdg );
        return true;
      }
      break;
  case XWMSG_KEY_PRESS:
      goto process_key;
  case WDGMSG_RECONFIGURE:
      My3DWidgetReshape ( wdg, x, y );
      break;
  case XWMSG_CLIENT_MESSAGE:
      goto process_client_message;
  default:
      break;
    }
    break;

case WDGSTATE_VIEW_TURNING:
    switch ( msg ) {
  case XWMSG_MOUSE_MOTION:
      if ( ((XMotionEvent*)wdg->wm->ev)->state & Button1Mask ) {
        if ( RotateViewer ( wd, (float)wdg->r.width, (float)wdg->r.height,
                            x - wd->prevx, y - wd->prevy ) ) {
          wd->prevx = x;  wd->prevy = y;
          wdg->wm->changed = true;
        }
      }
      else
        goto release;
      break;
  case XWMSG_BUTTON_RELEASE:
      if ( key == Button1 ) {
release:
        wdg->state = WDGSTATE_DEFAULT;
        UngrabInput ( wdg );
        return true;
      }
      break;
  case XWMSG_KEY_PRESS:
process_key:
      switch ( key ) {
    case ' ':
        ToggleAnimation ();
        return true;
    default:
        break;
      }
      break;
  case XWMSG_CLIENT_MESSAGE:
process_client_message:
      if ( ((XClientMessageEvent*)wdg->wm->ev)->message_type == aAnimate )
        AdvanceAnimation ();
  default:
      break;
    }
    break;

default:
    break;
  }
  return false;
} /*My3DWidgetInput*/

xwidget *New3DWidget ( struct xwinmenu *wm, int id, int w, int h )
{
  return NewWidget ( wm, sizeof(xwidget), id, w, h, 0, 0,
                     My3DWidgetInput, My3DWidgetRedraw, &myglwdata, NULL );
} /*New3DWidget*/

void RedrawWin2 ( xwinmenu *wm )
{
  xwidget *wdg;

  if ( opti > 0 ) {
    opti --;
    PostExposeEvent ( wm->window, wm->r.width, wm->r.height );
  }
  else {
    for ( wdg = wm->wlist.next;  wdg;  wdg = wdg->link.next )
      wdg->redraw ( wdg );
    glFlush ();
  }
  glXSwapBuffers ( xdisplay, wm->window );
} /*RedrawWin2*/

void Win2Reshape ( xwinmenu *wm, short w, short h )
{
  xwidget *wdg;

  for ( wdg = wm->wlist.next;  wdg;  wdg = wdg->link.next )
    wdg->input ( wdg, WDGMSG_RECONFIGURE, 0, w, h );
  opti = 3;
  PostExposeEvent ( wm->window, w, h );
  ExitIfGLError ( "Win2Reshape" );
} /*Win2Reshape*/

void Win2Callback ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  switch ( msg ) {
case WDGMSG_RECONFIGURE:
    Win2Reshape ( wdg->wm, x, y );
    break;
case XWMSG_CLIENT_MESSAGE:
    mywdg->input ( mywdg, msg, key, x, y );
    break;
default:
    break;
  }
} /*Win2Callback*/

xwinmenu *SetupApp3dGLWindow ( void )
{
  xwinmenu *wm;

  if ( !(wm = NewWinMenu ( window[2], WIN0_WIDTH-MENU1_WIDTH,
                           WIN0_HEIGHT-MENU3_HEIGHT, 0, 0,
                           NULL, RedrawWin2, Win2Callback )) )
    ExitOnError ( "SetupApp3dGLWindow 0" );
  if ( !(mywdg = New3DWidget ( wm, GLWIN_ID_VIEW, wm->r.width, wm->r.height )) )
    ExitOnError ( "SetupApp3dGLWindow 1" );
  return wm;
} /*SetupApp3dGLWindow*/

/* ////////////////////////////////////////////////////////////////////////// */
void Win3Reshape ( xwinmenu *wm, short w, short h )
{
  myknotwdg->wdg.input ( &myknotwdg->wdg, WDGMSG_RECONFIGURE, 0,
                     w-MENU1_WIDTH, MENU3_HEIGHT-1 );
  PostExposeEvent ( wm->window, w, h );
} /*Win3Reshape*/

void Win3Callback ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  switch ( msg ) {
case WDGMSG_RECONFIGURE:
    Win3Reshape ( wdg->wm, x, y );
    break;

case WDGMSG_BUTTON_PRESS:
    switch ( wdg->id ) {
  case BTN_ID_PLAY:
      ToggleAnimation ();
      break;
  case BTN_ID_LEFT:
      FindKeyFrame ( false );
      if ( myglwdata.animation )
        ToggleAnimation ();
      else
        goto let_redraw_it;
      break;
  case BTN_ID_RIGHT:
      FindKeyFrame ( true );
      if ( myglwdata.animation )
        ToggleAnimation ();
      else
        goto let_redraw_it;
      break;
  case BTN_ID_SET:
      SetKeyFrame ();
      break;
  default:
      break;
    }
    break;

case WDGMSG_SWITCH_CHANGE:
    switch ( wdg->id ) {
  case SW_ID_ANIMATE_KL:
      if ( !myscene.animate_kl && !myscene.animate_vp )
        myscene.animate_mm = true;
      break;
  case SW_ID_ANIMATE_MM:
      if ( !myscene.animate_mm && !myscene.animate_vp )
        myscene.animate_kl = true;
      break;
  case SW_ID_ANIMATE_VP:
      if ( !myscene.animate_vp && !myscene.animate_kl )
        myscene.animate_mm = true;
      break;
  case SW_ID_EDIT_KNOTS:
      if ( myknotwdg->editswitch )
        myknotwdg->panswitch = false;
      break;
  case SW_ID_KNW_PANZOOM:
      if ( myknotwdg->panswitch )
        myknotwdg->editswitch = false;
      break;
  default:
      break;
    }
    break;

case WDGMSG_KNOT_MCLICK:
    if ( key == 0 || key == 2 ) {
      KnotWidgetPoint ( x );
      goto let_redraw_it;
    }
    else
      break;

case WDGMSG_KNOT_MMOVE:
    KnotWidgetPoint ( x );
    goto let_redraw_it;

case WDGMSG_KNOT_CHANGE:
    ChangeKeyFrame ();
    goto let_redraw_it;

case WDGMSG_KNOT_INSERT:
    InsertKeyFrame ();
    goto let_redraw_it;

case WDGMSG_KNOT_DELETE:
    DeleteKeyFrame ();
let_redraw_it:
    wm1->changed = wm2->changed = wm3->changed = true;
    PostMenuExposeEvent ( wm1 );
    PostMenuExposeEvent ( wm2 );
    break;

default:
    break;
  }
} /*Win3Callback*/

void MyButtonRedraw ( struct xwidget *wdg )
{
  XRectangle rect[2];
  XPoint tr[3];

  XSetForeground ( xdisplay, xgc, XWP_BUTTON_COLOUR );
  XFillRectangle ( xdisplay, wdg->wm->pixmap, xgc,
                   wdg->r.x, wdg->r.y, wdg->r.width-1, wdg->r.height-1 );
  XSetForeground ( xdisplay, xgc, XWP_TEXT_COLOUR );
  XDrawRectangle ( xdisplay, wdg->wm->pixmap, xgc,
                   wdg->r.x, wdg->r.y, wdg->r.width-1, wdg->r.height-1 );
  if ( wdg->state == WDGSTATE_DEFAULT ) {
    tr[0].x = tr[1].x = wdg->r.x+wdg->r.width/2-4;  tr[2].x = wdg->r.x+wdg->r.width/2+4;
    tr[0].y = wdg->r.y+4;  tr[1].y = wdg->r.y+wdg->r.height-4;
    tr[2].y = wdg->r.y+wdg->r.height/2;
    XFillPolygon ( xdisplay, wdg->wm->pixmap, xgc, tr, 3, Convex, CoordModeOrigin );
  }
  else {
    rect[0].width = rect[1].width = 3;
    rect[0].height = rect[1].height = wdg->r.height-8;
    rect[0].y = rect[1].y = wdg->r.y+4;
    rect[0].x = wdg->r.x+wdg->r.width/2-4;  rect[1].x = wdg->r.x+wdg->r.width/2+3;
    XFillRectangles ( xdisplay, wdg->wm->pixmap, xgc, rect, 2 );
  }
} /*MyButtonRedraw*/

xwinmenu *SetupApp3dWin3Menu ( void )
{
  xwinmenu *wm;

  if ( !(wm = NewWinMenu ( window[3], WIN0_WIDTH, MENU3_HEIGHT,
                     0, WIN0_HEIGHT-MENU3_HEIGHT, NULL, NULL, Win3Callback)) )
    ExitOnError ( "SetupApp3dWin3Menu" );
  myknotwdg = NewKnotsWidget ( wm, KNOTWD_ID, WIN0_WIDTH-MENU1_WIDTH,
                 MENU3_HEIGHT-4, MENU1_WIDTH, 4, 4, MAXKEYFRAMES,
                 myscene.lastkeyframe+1, myscene.keyknots,
                 myscene.keyknots[0], myscene.keyknots[3] );
  myknotwdg->editswitch = true;
  myknotwdg->panswitch = false;
  NewButton ( wm, BTN_ID_LEFT, 16, 18, 0, 0, str_LEFT );
  NewButton ( wm, BTN_ID_SET, 22, 18, 19, 0, str_SET );
  NewButton ( wm, BTN_ID_RIGHT, 16, 18, 44, 0, str_RIGHT );
  playbtn = NewButton ( wm, BTN_ID_PLAY, 60, 18, 0, 20, NULL );
  playbtn->redraw = MyButtonRedraw;
  NewSwitch ( wm, SW_ID_ANIMATE_KL, 16, 16, 63, 20, NULL, &myscene.animate_kl );
  NewSwitch ( wm, SW_ID_ANIMATE_MM, 16, 16, 82, 20, NULL, &myscene.animate_mm );
  NewSwitch ( wm, SW_ID_ANIMATE_VP, 16, 16, 101, 20, NULL, &myscene.animate_vp );
  NewSwitch ( wm, SW_ID_EDIT_KNOTS, 16, 16, 0, 40, str_EDIT,
              &myknotwdg->editswitch );
  NewSwitch ( wm, SW_ID_KNW_PANZOOM, 16, 16, 44, 40, str_PANZOOM,
              &myknotwdg->panswitch );
  return wm;
} /*SetupApp3dWin3Menu*/

/* ////////////////////////////////////////////////////////////////////////// */
void MessageLoop ( void )
{
  XEvent ev;

  terminate = false;
  do {
    XNextEvent ( xdisplay, &ev );
    if ( ev.xany.window == window[0] )
      Win0MessageProc ( &ev );
    else if ( ev.xany.window == window[1] )
      WinMenuInput ( wm1, &ev );
    else if ( ev.xany.window == window[2] )
      WinMenuInput ( wm2, &ev );
    else if ( ev.xany.window == window[3] )
      WinMenuInput ( wm3, &ev );
  } while ( !terminate );
} /*MessageLoop*/

void InitApp3dWindows ( int argc, char **argv, int major, int minor )
{
  int visattr[] =
    { GLX_RGBA,
      GLX_DOUBLEBUFFER,
      GLX_RED_SIZE,    8,
      GLX_GREEN_SIZE,  8,
      GLX_BLUE_SIZE,   8,
      GLX_DEPTH_SIZE, 24,
      GLX_SAMPLE_BUFFERS, True,
      GLX_SAMPLES,     8,
      None };
  XSetWindowAttributes swa;
  static const int     wx[4] = { 0, 0, MENU1_WIDTH, 0 };
  static const int     wy[4] = { 0, 0, 0, WIN0_HEIGHT-MENU3_HEIGHT };
  static const int     ww[4] = { WIN0_WIDTH, MENU1_WIDTH,
                                 WIN0_WIDTH-MENU1_WIDTH, WIN0_WIDTH };
  static const int     wh[4] = { WIN0_HEIGHT, WIN0_HEIGHT-MENU3_HEIGHT,
                                 WIN0_HEIGHT-MENU3_HEIGHT, MENU3_HEIGHT };
  Window               upw;
  int                  i;

  if ( !(xdisplay = XOpenDisplay ( "" )) )
    ExitOnError ( "InitApp3dWindows 0" );
  xscreen = DefaultScreen ( xdisplay );
  xrootwin = RootWindow ( xdisplay, xscreen );
  InitGLXContext ( major, minor, visattr );
  if ( !(xcolormap = XCreateColormap ( xdisplay, xrootwin,
                                       xvisual, AllocNone )) )
    ExitOnError ( "InitApp3dWindows 1" );
  swa.colormap = xcolormap;
  swa.event_mask = ExposureMask | StructureNotifyMask| ButtonPressMask |
                   ButtonReleaseMask | PointerMotionMask | KeyPressMask |
                   EnterWindowMask | LeaveWindowMask ;
  for ( i = 0, upw = xrootwin; i < 4; i++, upw = window[0] ) {
    window[i] = XCreateWindow ( xdisplay, upw, wx[i], wy[i], ww[i], wh[i],
                             0, 24, InputOutput, xvisual,
                             CWColormap | CWEventMask, &swa );
    XMapWindow ( xdisplay, window[i] );
  }
  XStoreName ( xdisplay, window[0], "Aplikacja 3D" );
  xgc = XCreateGC ( xdisplay, window[1], 0, 0 );
  aAnimate = XInternAtom ( xdisplay, "aAnimate", False );
  InitRGBXColourmap (); 
  InitWinMenuPalette ();
  if ( !glXMakeCurrent ( xdisplay, window[2], glxcontext ) )
    ExitOnError ( "InitApp3dWindows 2" );
  GetGLProcAddresses ();
  window0_width = WIN0_WIDTH;  
  window0_height = WIN0_HEIGHT;
} /*InitApp3dWindows*/

void Initialise ( int argc, char **argv )
{
  InitApp3dWindows ( argc, argv, APP3_MAJOR, APP3_MINOR );
  LoadMyShaders ();
  InitMyObject ();
  wm1 = SetupApp3dMenu ();
  wm2 = SetupApp3dGLWindow ();
  wm3 = SetupApp3dWin3Menu ();
  Win2Reshape ( wm2, wm2->r.width, wm2->r.height );
} /*Initialise*/

void Cleanup ( void )
{
  DestroyMyObject ();
  DestroyMyShaders ();
  DestroyWinMenu ( wm1 );
  DestroyWinMenu ( wm2 );
  DestroyWinMenu ( wm3 );
  XDestroySubwindows ( xdisplay, window[0] );
  XDestroyWindow ( xdisplay, window[0] );
  XFreeGC ( xdisplay, xgc );
  XFree ( glxfbc );  
  XCloseDisplay ( xdisplay );
} /*Cleanup*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/
